#! /usr/bin/env python
"""Horrible hack to attempt to load meliae dumps a bit faster

Makes meliae loading about 4.25x faster on python 2.6 compared to the 
json + C speedups.  This is *not* however, a full json decoder, it is 
*just* a parser for the flat records meliae produces (i.e. no recursive 
structures, no floats, just ints, strings and lists-of-ints)
"""
import re, unittest, json

whitespace = r'[ \t]'

escape = r"""(?:\\[uU][0-9a-fA-F]{4})"""
string = r"""(?:["](?P<%(key)s>([^"\\]|(\\")|%(escape)s|\\[^uU"])*?)["])"""
key = string%{'key':'key','escape':escape}
string = string%{'key':'string','escape':escape}
integer = r"""(?P<int>[+-]*\d+)"""
listcontent = r"""([+-]*\d+[,]?%(whitespace)s*?)*"""%globals()
intlist = r"""\[%(whitespace)s*(?P<list>%(listcontent)s)%(whitespace)s*\]"""%globals()

attr = r"""%(whitespace)s*%(key)s%(whitespace)s*:%(whitespace)s*(%(intlist)s|%(string)s|%(integer)s)(,)?"""%globals()

escape = re.compile( escape, re.U )
simple_escape = re.compile( r'\\([^uU])', re.U )

assert escape.match( "\\u0000" )
attr = re.compile( attr )
string = re.compile( string )
integer = re.compile( integer )
intlist = re.compile( intlist )

assert string.match( '"this"' ).group('string') == "this"
assert string.match( '"this": "that"' ).group('string') == "this"
assert string.match( '"this\\u0000"' ).group('string') == "this\\u0000", string.match( '"this\\u0000"' ).group('string')

assert integer.match( '23' ).group( 'int' ) == '23'
assert intlist.match( '[1, 2,3,4]' ).group( 'list' ) == '1, 2,3,4'
assert intlist.match( '[139828625414688, 70572696, 52870672, 40989336]' ).group('list') == '139828625414688, 70572696, 52870672, 40989336'

def loads( source ):
    """Load json structure from meliae from source
    
    Supports only the required structures to support loading meliae memory dumps
    """
    source = source.strip()
    assert source.startswith( '{' )
    assert source.endswith( '}' )
    source = source[1:-1]
    result = {}
    for match in attr.finditer( source ):
        key = match.group('key')
        if match.group( 'list' ) is not None:
            value = [ 
                int(x) 
                for x in match.group( 'list' ).strip().replace(',',' ').split() 
            ]
        elif match.group( 'int' ) is not None:
            value = int( match.group( 'int' ))
        elif match.group( 'string' ) is not None:
            def deescape( match ):
                return chr( int( match.group(0)[2:], 16 ))
            value = match.group('string').decode( 'utf-8' )
            value = escape.sub( 
                deescape,
                value,
            )
            value = simple_escape.sub(
                lambda x: x.group(1),
                value,
            )
        else:
            raise RuntimeError( "Matched something we don't know how to process:", match.groupdict() )
        result[key] = value
    return result
if __name__ == "__main__":
    import sys, pprint
    for line in open( sys.argv[1] ):
        official = json.loads( line )



import unittest
from profile_viewer import meliaeloader

def records_as_index( records ):
    index = dict([
        (x['address'],x)
        for x in records 
    ])
    shared = {}
    for value in records:
        for ref in value['refs']:
            shared.setdefault( ref, [] ).append( value['address'] )
    return index, shared

class MeliaeTests( unittest.TestCase ):
    def test_simplify_dict( self ):
        """Do we remove __dict__ items properly?"""
        records = [
            {'size': 10,'type':'moo','address':1,'refs':[]},
            {'size': 1,'type':'dict','address':2,'refs':[1]},
            {'size': 1,'type':'module','address':3,'refs':[2]},
        ]
        index,shared = records_as_index( records )
        meliaeloader.bind_parents( index, shared )
        meliaeloader.simplify_dicts( index, shared )
        assert index == {
            3:{'size': 2,'type':'module','address':3,'refs':[1], 'parents':[],'compressed':True},
            1:{'size': 10,'type':'moo','address':1,'refs':[],'parents':[3]},
        }, index
    
    def test_group_children( self ):
        """Do we group large numbers of children properly?"""
        strings = [
            {'size':1,'type':'str','address': i+12,'refs':[]}
            for i in range( 10 )
        ]
        records = [
            {'size': 10,'type':'moo','address':1,'refs':[s['address'] for s in strings]},
            {'size': 1,'type':'dict','address':2,'refs':[1]},
            {'size': 1,'type':'module','address':3,'refs':[2]},
        ] + strings
        index,shared = records_as_index( records )
        meliaeloader.group_children( index, shared )
        assert len(index) == 4, index # module, dict, moo, collection-of-strings
        
        new = index[-1]
        assert len(new['refs']) == 0, new
        assert new['type'] == '<many>', new
        assert new['name'] == 'str', new
        assert len(new['parents']) == 1, new

    def test_group_module_children( self ):
        """Do we group large numbers of children properly?"""
        strings = [
            {'size':1,'type':'str','address': i+12,'refs':[]}
            for i in range( 10 )
        ]
        records = [
            {'size': 1,'type':'dict','address':2,'refs':[s['address'] for s in strings]},
            {'size': 1,'type':'module','address':3,'refs':[2]},
        ] + strings
        index,shared = records_as_index( records )
        meliaeloader.bind_parents( index, shared )
        meliaeloader.group_children( index, shared )
        meliaeloader.simplify_dicts( index, shared )
        assert len(index) == 2, index # module, collection-of-strings
        
        new = index[-1]
        assert len(new['refs']) == 0, new
        assert new['type'] == '<many>', new
        assert new['name'] == 'str', new
        assert len(new['parents']) == 1, new
        assert new['parents'] == [3], new

    def test_recursive( self ):
        """Do we account for recursive structures properly?"""
        records = [
            {'size': 10,'type':'moo','address':1,'refs':[4]},
            {'size': 10,'type':'moo','address':4,'refs':[1]},
            {'size': 0,'type':'moo','address':5,'refs':[1,4]},
            {'size': 1,'type':'dict','address':2,'refs':[5]},
            {'size': 1,'type':'module','address':3,'refs':[2]},
        ]
        index,shared = records_as_index( records )
        meliaeloader.bind_parents( index, shared )
        meliaeloader.simplify_dicts( index, shared )
        loops = list( meliaeloader.find_loops( index[3], index ) )
        assert loops
        meliaeloader.promote_loops( loops, index, shared )
        
        meliaeloader.recurse_module( 
            index[3], 
            index, 
            shared,
        )
        assert index[3]['totsize'] == 22, index
    
    def test_recursive_shared( self ):
        """Do we account for recursive structures shared across modules properly?
        
        Loops should become a single "thing" which has references and parents, but 
        also can be broken down to see what is in the loop...
        
        Loop totsize is the totsize of all components of the loop divided by the 
        number of references to the loop which are *not* from the loop itself...
        """
        records = [
            {'size': 10,'type':'moo','address':1,'refs':[4]},
            {'size': 10,'type':'moo','address':4,'refs':[1]},
            {'size': 0,'type':'moo','address':7,'refs':[1]},
            {'size': 1,'type':'dict','address':2,'refs':[7]},
            {'size': 1,'type':'module','address':3,'refs':[2]},
            
            {'size': 1,'type':'dict','address':5,'refs':[4,1]}, # should become a single ref to loop...
            {'size': 1,'type':'module','address':6,'refs':[5]},
        ]
        index,shared = records_as_index( records )
        meliaeloader.bind_parents( index, shared )
        meliaeloader.simplify_dicts( index, shared )
        
        loops = list( meliaeloader.find_loops( index[6], index ) )
        assert len(loops) == 1, loops 
        loop = loops[0]
        assert set(loop) == set([1,4]), loop
        
        meliaeloader.promote_loops( loops, index, shared )
        
        for module in [3,6]:
            meliaeloader.recurse_module( 
                index[module], 
                index=index, 
                shared=shared,
            )
            assert index[module]['totsize'] == 12, index[module]['totsize']

        loop = [x for x in list(index.values()) if x['type'] == '<loop>'][0]
        assert set(loop['parents']) == set([6,7]), loop['parents']
    



#! /usr/bin/env python
"""Module to load meliae memory-profile dumps

Trees:

    * has-a
        * module root 
        * each held reference contributes a weighted cost to the parent 
        * hierarchy of held objects, so globals, classes, functions, and their children
        * held modules do not contribute to cost
        
        * module 
            * instance-tree
"""
import logging, sys, weakref
log = logging.getLogger( __name__ )
from gettext import gettext as _
try:
    from ._meliaejson import loads as json_loads
except ImportError as err:
    try:
        from json import loads as json_loads
    except ImportError as err:
        from simplejson import loads as json_loads
import sys

LOOP_TYPE = _('<loop>')
MANY_TYPE = _('<many>')
NON_MODULE_REFS = _('<non-module-references>')

STOP_TYPES = set(['module'])

def recurse( record, index, stop_types=STOP_TYPES,already_seen=None, type_group=False ):
    """Depth first traversal of a tree, all children are yielded before parent
    
    record -- dictionary record to be recursed upon 
    index -- mapping 'address' ids to dictionary records 
    stop_types -- types which will *not* recurse 
    already_seen -- set storing already-visited nodes 
    
    yields the traversed nodes
    """
    if already_seen is None:
        already_seen = set()
    if record['address'] not in already_seen:
        already_seen.add(record['address'])
        if 'refs' in record:
            for child in children( record, index, stop_types=stop_types ):
                if child['address'] not in already_seen:
                    for descendant in recurse( 
                        child, index, stop_types, 
                        already_seen=already_seen, type_group=type_group,
                    ):
                        yield descendant
        yield record 

def find_loops( record, index, stop_types = STOP_TYPES, open=None, seen = None ):
    """Find all loops within the index and replace with loop records"""
    if open is None:
        open = []
    if seen is None:
        seen = set()
    for child in children( record, index, stop_types = stop_types ):
        if child['type'] in stop_types or child['type'] == LOOP_TYPE:
            continue
        if child['address'] in open:
            # loop has been found 
            start = open.index( child['address'] )
            new = frozenset( open[start:] )
            if new not in seen:
                seen.add(new)
                yield new
        elif child['address'] in seen:
            continue 
        else:
            seen.add( child['address'])
            open.append( child['address'] )
            for loop in find_loops( child, index, stop_types=stop_types, open=open, seen=seen ):
                yield loop 
            open.pop( -1 )

def promote_loops( loops, index, shared ):
    """Turn loops into "objects" that can be processed normally"""
    for loop in loops:
        loop = list(loop)
        members = [index[addr] for addr in loop]
        external_parents = list(set([
            addr for addr in sum([shared.get(addr,[]) for addr in loop],[])
            if addr not in loop 
        ]))
        if external_parents:
            if len(external_parents) == 1:
                # potentially a loop that's been looped...
                parent = index.get( external_parents[0] )
                if parent['type'] == LOOP_TYPE:
                    continue 
            # we haven't already been looped...
            loop_addr = new_address( index )
            shared[loop_addr] = external_parents
            loop_record = index[loop_addr] = {
                'address': loop_addr,
                'refs': loop,
                'parents': external_parents,
                'type': LOOP_TYPE,
                'size': 0,
            }
            for member in members:
                # member's references must *not* point to loop...
                member['refs'] = [
                    ref for ref in member['refs']
                    if ref not in loop 
                ]
                # member's parents are *just* the loop
                member['parents'][:] = [loop_addr]
            # each referent to loop holds a single reference to the loop rather than many to children
            for parent in external_parents:
                parent = index[parent]
                for member in members:
                    rewrite_references( parent['refs'], member['address'], None )
                parent['refs'].append( loop_addr )

def children( record, index, key='refs', stop_types=STOP_TYPES ):
    """Retrieve children records for given record"""
    result = []
    for ref in record.get( key,[]):
        try:
            record = index[ref]
        except KeyError as err:
            #print 'No record for %s address %s in %s'%(key, ref, record['address'] )
            pass # happens when an unreachable references a reachable that has been compressed out...
        else:
            if record['type'] not in stop_types:
                result.append(  record  )
    return result

def children_types( record, index, key='refs', stop_types=STOP_TYPES ):
    """Produce dictionary mapping type-key to instances for all children"""
    types = {}
    for child in children( record, index, key, stop_types=stop_types ):
        types.setdefault(child['type'],[]).append( child )
    return types
        

def recurse_module( overall_record, index, shared, stop_types=STOP_TYPES, already_seen=None, min_size=0 ):
    """Creates a has-a recursive-cost hierarchy
    
    Mutates objects in-place to produce a hierarchy of memory usage based on 
    reference-holding cost assignment
    """
    for record in recurse( 
        overall_record, index, 
        stop_types=stop_types, 
        already_seen=already_seen, 
        type_group=True,
    ):
        # anything with a totsize we've already processed...
        if record.get('totsize') is not None:
            continue 
        rinfo = record 
        rinfo['module'] = overall_record.get('name',NON_MODULE_REFS )
        if not record['refs']:
            rinfo['rsize'] = 0
            rinfo['children'] = []
        else:
            # TODO: provide a flag to coalesce based on e.g. type at each level or throughout...
            rinfo['children'] = rinfo_children = list ( children( record, index, stop_types=stop_types ) )
            rinfo['rsize'] = sum([
                (
                    child.get('totsize',0.0)/float(len(shared.get( child['address'], [])) or 1)
                )
                for child in rinfo_children
            ], 0.0 )
        rinfo['totsize'] = record['size'] + rinfo['rsize']
    
    return None
    
def as_id( x ):
    if isinstance( x, dict ):
        return x['address']
    else:
        return x

def rewrite_refs( targets, old,new, index, key='refs', single_ref=False ):
    """Rewrite key in all targets (from index if necessary) to replace old with new"""
    for parent in targets:
        if not isinstance( parent, dict ):
            try:
                parent = index[parent]
            except KeyError as err:
                continue 
        rewrite_references( parent[key], old, new, single_ref=single_ref )

def rewrite_references( sequence, old, new, single_ref=False ):
    """Rewrite parents to point to new in old
    
    sequence -- sequence of id references 
    old -- old id 
    new -- new id
    
    returns rewritten sequence
    """
    old,new = as_id(old),as_id(new)
    to_delete = []
    for i,n in enumerate(sequence):
        if n == old:
            if new is None:
                to_delete.append( i )
            else:
                sequence[i] = new 
                if single_ref:
                    new = None
        elif n == new and single_ref:
            new = None
    if to_delete:
        to_delete.reverse()
        for i in to_delete:
            del sequence[i]
    return sequence

def simple( child, shared, parent ):
    """Return sub-set of children who are "simple" in the sense of group_children"""
    return (
        not child.get('refs',())
        and (
            not shared.get(child['address'])
        or 
            shared.get(child['address']) == [parent['address']]
        )
    )

def group_children( index, shared, min_kids=10, stop_types=STOP_TYPES, delete_children=True ):
    """Collect like-type children into sub-groups of objects for objects with long children-lists
    
    Only group if:
    
        * there are more than X children of type Y
        * children are "simple"
            * individual children have no children themselves
            * individual children have no other parents...
    """
    to_compress = []
    
    for to_simplify in list(iterindex( index )):
        if not isinstance( to_simplify, dict ):
            continue
        for typ,kids in list(children_types( to_simplify, index, stop_types=stop_types ).items()):
            kids = [k for k in kids if k and simple(k,shared, to_simplify)]
            if len(kids) >= min_kids:
                # we can group and compress out...
                to_compress.append( (to_simplify,typ,kids))
    
    for to_simplify,typ,kids in to_compress:
        typ_address = new_address(index)
        kid_addresses = [k['address'] for k in kids]
        index[typ_address] = {
            'address': typ_address,
            'type': MANY_TYPE,
            'name': typ,
            'size': sum( [k.get('size',0) for k in kids], 0),
            'parents': [to_simplify['address']],
        }
        
        shared[typ_address] = index[typ_address]['parents']
        to_simplify['refs'][:] = [typ_address]
        
        if delete_children:
            for address in kid_addresses:
                try:
                    del index[address]
                except KeyError as err: 
                    pass # already compressed out
                try:
                    del shared[address]
                except KeyError as err:
                    pass # already compressed out
            index[typ_address]['refs'] = []
        else:
            index[typ_address]['refs'] = kid_addresses

# Types which *can* have their dictionaries compressed out
SIMPLIFY_DICTS = set( ['module','type','classobj'])
# Types which will *always* have their dictionaries compressed out,
# even if there are multiple references to the dictionary, these values 
# should *only* ever be part of STOP_TYPES, as their size contributions 
# will be lost (STOP_TYPES make no contribution)
ALWAYS_COMPRESS_DICTS = set( ['module'] )


def simplify_dicts( index, shared, simplify_dicts=SIMPLIFY_DICTS, always_compress=ALWAYS_COMPRESS_DICTS ):
    """Eliminate "noise" dictionary records from the index 
    
    index -- overall index of objects (including metadata such as type records)
    shared -- parent-count mapping for records in index
    
    module/type/class dictionaries
    """
    
    # things which will have their dictionaries compressed out
    
    to_delete = set()
    
    for to_simplify in iterindex(index):
        if to_simplify['address'] in to_delete:
            continue 
        if to_simplify['type'] in simplify_dicts and not 'compressed' in to_simplify:
            refs = to_simplify['refs']
            for ref in refs:
                child = index.get( ref )
                if child is not None and child['type'] == 'dict':
                    child_referrers = child['parents'][:]
                    if len(child_referrers) == 1 or to_simplify['type'] in always_compress:
                        
                        to_simplify['compressed'] = True
                        to_simplify['refs'] = child['refs']
                        to_simplify['size'] += child['size']
                        
                        # rewrite anything *else* that was pointing to child to point to us...
                        while to_simplify['address'] in child_referrers:
                            child_referrers.remove( to_simplify['address'] )
                        if child_referrers:
                            rewrite_refs( 
                                child_referrers, 
                                child['address'],
                                to_simplify['address'], 
                                index, single_ref=True
                            )
                        
                        # now rewrite grandchildren to point to root obj instead of dict
                        for grandchild in child['refs']:
                            grandchild = index[grandchild]
                            parent_set = grandchild['parents']
                            if parent_set:
                                rewrite_references( 
                                    parent_set, 
                                    child,
                                    to_simplify,
                                    single_ref = True,
                                )
                            assert parent_set
                        to_delete.add( child['address'] )
    for item in to_delete:
        del index[item]
        del shared[item]
    
    return index

def find_reachable( modules, index, shared, stop_types=STOP_TYPES ):
    """Find the set of all reachable objects from given root nodes (modules)"""
    reachable = set()
    already_seen = set()
    for module in modules:
        for child in recurse( module, index, stop_types=stop_types, already_seen=already_seen):
            reachable.add( child['address'] )
    return reachable

def deparent_unreachable( reachable, shared ):
    """Eliminate all parent-links from unreachable objects from reachable objects
    """
    for id,shares in shared.items():
        if id in reachable: # child is reachable
            filtered = [
                x 
                for x in shares 
                if x in reachable # only those parents which are reachable
            ]
            if len(filtered) != len(shares):
                shares[:] = filtered

class _syntheticaddress( object ):
    current = -1
    def __call__( self, target ):
        while self.current in target:
            self.current -= 1
        target[self.current] = True
        return self.current 
new_address = _syntheticaddress()

def index_size( index ):
    return sum([
        v.get('size',0)
        for v in iterindex( index )
    ],0)

def iterindex( index ):
    for (k,v) in index.items():
        if (
            isinstance(v,dict) and 
            isinstance(k,int)
        ):
            yield v

def bind_parents( index, shared ):
    """Set parents on all items in index"""
    for v in iterindex( index ):
        v['parents'] = shared.get( v['address'], [] )


def check_parents( index, reachable ):
    for item in iterindex( index ):
        if item['type'] == '<many>':
            print('parents', item['parents'])

def load( filename, include_interpreter=False ):
    index = {
    } # address: structure
    shared = dict() # address: [parent addresses,...]
    modules = set()
    
    root_address = new_address( index )
    root = {
        'type':'dump',
        'name': filename,
        'children': [],
        'totsize': 0,
        'rsize': 0,
        'size': 0,
        'address': root_address,
    }
    index[root_address] = root
    index_ref = Ref( index )
    root_ref = Ref( root )
    
    root['root'] = root_ref 
    root['index'] = index_ref
    
    raw_total = 0
    
    for line in open( filename ):
        struct = json_loads( line.strip())
        index[struct['address']] = struct 
        
        struct['root'] = root_ref
        struct['index'] = index_ref

        refs = struct['refs']
        for ref in refs:
            parents = shared.get( ref )
            if parents is None:
                shared[ref] = []
            shared[ref].append( struct['address'])
        raw_total += struct['size']
        if struct['type'] == 'module':
            modules.add( struct['address'] )
    
    modules = [index[addr] for addr in modules]
    
    reachable = find_reachable( modules, index, shared )
    deparent_unreachable( reachable, shared )
    
    bind_parents( index, shared )
    
#    unreachable = sum([
#        v.get( 'size' )
#        for v in iterindex( index )
#        if v['address'] not in reachable
#    ], 0 )
#    print '%s bytes are unreachable from modules'%( unreachable )

    simplify_dicts( index,shared )

    group_children( index, shared, min_kids=10 )

    records = []
    for m in modules:
        loops = list( find_loops( m, index ) )
        promote_loops( loops, index, shared )
        recurse_module(
            m, index, shared
        )
        
    modules.sort( key = lambda m: m.get('totsize',0))
    for module in modules:
        module['parents'].append( root_address )


    if include_interpreter:
        # Meliae produces quite a few of these un-referenced records, they aren't normally useful AFAICS
        # reachable from any module, but are present in the dump...
        disconnected = [
            x for x in iterindex( index )
            if x.get('totsize') is None 
        ]
        for pseudo_module in find_roots( disconnected, index, shared ):
            pseudo_module['root'] = root_ref
            pseudo_module['index'] = index_ref 
            pseudo_module.setdefault('parents',[]).append( root_address )
            modules.append( pseudo_module )
    else:
        to_delete = []
        for v in iterindex(index):
            if v.get('totsize') is None:
                to_delete.append( v['address'] )
        for k in to_delete:
            del index[k]

    all_modules = sum([x.get('totsize',0) for x in modules],0)
 
    root['totsize'] = all_modules
    root['rsize'] = all_modules
    root['size'] = 0
    root['children'] = modules

    for item in iterindex( index ):
        item['root'] = root_ref
        item['index'] = index_ref
    
    return root, index

def find_roots( disconnected, index, shared ):
    """Find appropriate "root" objects from which to recurse the hierarchies
    
    Will generate a synthetic root for anything which doesn't have any parents...
    """
    log.warn( '%s disconnected objects in %s total objects', len(disconnected), len(index))
    natural_roots = [x for x in disconnected if x.get('refs') and not x.get('parents')]
    log.warn( '%s objects with no parents at all' ,len(natural_roots))
    for natural_root in natural_roots:
        recurse_module(
            natural_root, index, shared
        )
        yield natural_root
    rest = [x for x in disconnected if x.get( 'totsize' ) is None]
    un_found = {
        'type': 'module',
        'name': '<disconnected objects>',
        'children': rest,
        'parents': [ ],
        'size': 0,
        'totsize': sum([x['size'] for x in rest],0),
        'address': new_address( index ),
    }
    index[un_found['address']] = un_found
    yield un_found

class Ref(object):
    def __init__( self, target ):
        self.target = target
    def __call__( self ):
        return self.target

class Loader( object ):
    """A data-set loader for pulling root and rows from a meliae dump"""
    def __init__( self, filename, include_interpreter=False ):
        self.filename = filename 
        self.include_interpreter = include_interpreter
        self.roots = {}
    ROOTS = ['memory']
    
    def get_root( self, key ):
        """Retrieve the given root by type-key"""
        if key not in self.roots:
            root,self.rows = load( self.filename, include_interpreter = self.include_interpreter )
            self.roots[key] = root
        return self.roots[key]
    def get_rows( self, key ):
        """Get the set of rows for the type-key"""
        if key not in self.roots:
            self.get_root( key )
        return self.rows
    def get_adapter( self, key ):
        from profile_viewer import meliaeadapter
        return meliaeadapter.MeliaeAdapter()

if __name__ == "__main__":
    import logging
    logging.basicConfig( level=logging.DEBUG )
    import sys
    load( sys.argv[1] )
#    import cProfile, sys
#    cProfile.runctx( "load(sys.argv[1])", globals(),locals(),'melialoader.profile' )
    



"""Adapter for RunSnakeRun to load coldshot profiles"""
import wx, sys, os, logging
log = logging.getLogger( __name__ )
from squaremap import squaremap
from coldshot import stack,loader

class BaseColdshotAdapter( squaremap.DefaultAdapter):
    """Base class for the various adapters"""
    percentageView = False
    total = 0
    def filename( self, node ):
        return getattr(node,'path',None)
    color_mapping = None

    def background_color(self, node, depth):
        """Create a (unique-ish) background color for each node"""
        if self.color_mapping is None:
            self.color_mapping = {}
        color = self.color_mapping.get(node.key)
        if color is None:
            depth = len(self.color_mapping)
            red = (depth * 10) % 255
            green = 200 - ((depth * 5) % 200)
            blue = (depth * 25) % 200
            self.color_mapping[node.key] = color = wx.Colour(red, green, blue)
        return color

    def SetPercentage(self, percent, total):
        """Set whether to display percentage values (and total for doing so)"""
        self.percentageView = percent
        self.total = total

    def parents(self, node):
        return getattr(node, 'parents', [])
    def label(self, node):
        if self.percentageView and self.total:
            time = '%0.2f%%' % round(node.cumulative * 100.0 / self.total, 2)
        else:
            time = '%0.3fs' % round(node.cumulative, 3)
        if hasattr( node, 'line' ):
            return '%s@%s:%s [%s]' % (node.name, node.filename, node.line, time)
        else:
            return '%s [%s]'%( node.name, time )

class ColdshotAdapter(BaseColdshotAdapter):
    """Adapts a coldshot.loader.Loader into a Squaremap-compatible structure"""

    def value(self, node, parent=None):
        if parent:
            return parent.child_cumulative_time(node)
        else:
            return node.cumulative
    
    def empty(self, node):
        """Calculate percentage of "empty" time"""
        return node.empty

#
#class ColdshotCallsAdapter( BaseColdshotAdapter ):
#    def value(self, node, parent=None):
#        return node.cumulative / parent.cumulative
#    
#    def empty(self, node):
#        """Calculate percentage of "empty" time"""
#        return node.empty

class FunctionLineWrapper( object ):
    def __init__( self, function_info, line_info ):
        self.function_info = function_info
        self.line_info = line_info
    @property 
    def children( self ):
        return []
    @property 
    def parents( self ):
        return [ self.function_info ]
    @property 
    def cumulative( self ):
        return self.line_info.time * self.function_info.loader.timer_unit
    @property 
    def empty( self ):
        return 0.0
    @property 
    def local( self ):
        return self.line_info.time * self.function_info.loader.timer_unit
    @property 
    def key( self ):
        return self.function_info.key 
    @property 
    def name( self ):
        return '%s:%s'%( self.line_info.line, self.function_info.filename,  )
    @property 
    def calls( self ):
        return self.line_info.calls

class ModuleAdapter( ColdshotAdapter ):
    """Currently doesn't do anything different"""
    def label(self, node):
        if isinstance( node, stack.FunctionInfo ):
            return super( ModuleAdapter, self ).label( node )
        if self.percentageView and self.total:
            time = '%0.2f%%' % round(node.cumulative * 100.0 / self.total, 2)
        else:
            time = '%0.3fs' % round(node.cumulative, 3)
        return '%s [%s]'%(node.key or 'PYTHONPATH', time)
    def parents( self, node ):
        if isinstance( node, stack.FunctionInfo ):
            parent = node.loader.modules.get( node.module )
            if parent:
                return [parent]
            return []
        elif isinstance( node, stack.FunctionLineInfo ):
            return [node.function]
        else:
            return getattr( node, 'parents', [] )
    def children( self, node ):
        if isinstance( node, stack.FunctionInfo ):
            return [
                FunctionLineWrapper( node, line )
                for lineno,line in sorted( node.line_map.items())
            ]
        return ColdshotAdapter.children( self, node )
    def label(self, node):
        if isinstance( node, FunctionLineWrapper ):
            return node.name 
        return ColdshotAdapter.label( self, node )
    
        
class Loader( loader.Loader ):
    """Coldshot loader subclass with knowledge of squaremap adapters"""
    def functions_rows( self ):
        """Get cProfile-like function metadata rows
        
        returns an ID: function mapping
        """
        return self.info.functions
    def location_rows( self ):
        """Get our location records (finalized)
        
        returns an module-name: Grouping mapping
        """
        self.info.finalize_modules()
        return self.info.modules
        
    ROOTS = ['functions','location' ]# ,'thread','calls']
    
    def get_root( self, key ):
        """Retrieve the given root by type-key"""
        return self.info.roots[key]
    def get_rows( self, key ):
        """Get the set of rows for the type-key"""
        return getattr( self, '%s_rows'%(key,) )( )
    def get_adapter( self, key ):
        """Get an adapter for our given key"""
        if key == 'functions':
            return ColdshotAdapter()
        elif key == 'location':
            return ModuleAdapter()
        else:
            raise KeyError( """Unknown root type %s"""%( key, ))
    



#! /usr/bin/env python
import sys,http.client

class Test( object ):
    def __init__( self, size=1024 ):
        self.data = 'a'*size

# set up some reference loops
t = Test()
t2 = Test()
t3 = Test()
t2.first = t
t3.first = t
t.second = t 
t.third = t3

big_list = [ Test(i) for i in range(256) ]
big_list.append( big_list )

def main():

    from meliae import scanner
    filename = 'dump.memory'
    fh = open( filename, 'wb' )
    scanner.dump_all_objects( fh )
    fh.close()
    print('saved memory dump to: %r'%( filename, ))

if __name__ == "__main__":
    main()



#! /usr/bin/env python
"""Creates hotshot and cProfile sample files"""
import time,cProfile,hotshot
from subpackage.timewaster import r

def x( ):
    print('x')
    y()
    z()
    import big_import
    a()
    r()
def y( ):
    print('y')
    for i in range( 2500 ):
        int(i) ** i
    time.sleep( 0.25 )
    z()
def z( ):
    print('z')
    time.sleep( 0.1 )
    a()

def a( count=5 ):
    print('a',count)
    if count:
        time.sleep( 0.05 )
        return a( count - 1 )


if __name__ == "__main__":
    import pprint 
    command = '''x()'''
    profiler = hotshot.Profile( 
        "hotshot.profile", lineevents=True, linetimings=True 
    )
    profiler.runctx( command, globals(), locals())
    print(dir(profiler))
    profiler.close()
    print('hotshot line events' ,profiler.lineevents)
    
    profiler = cProfile.Profile( subcalls=True )
    profiler.runctx( command, globals(), locals())
    stats = profiler.getstats()
    profiler.dump_stats( 'cprofile.profile' )
    
    try:
        import line_profiler
    except ImportError as err:
        pass 
    else:
        profiler  = line_profiler.LineProfiler()
#        profiler.add_function( x )
        for func in (a,x,y,z):
            profiler.add_function( func )
        profiler.runctx( command, globals(), locals() )
        profiler.dump_stats( 'line_profiler.profile' )
        
    



from setuptools import setup, find_packages

setup(
    name="profile-viewer",
    url="http://www.vrplumber.com/programming/profile_viewer/",
    download_url="http://www.vrplumber.com/programming/profile_viewer/",
    description="GUI Viewer for Python profiling runs",
    author="Mike C. Fletcher",
    author_email="mcfletch@vrplumber.com",
    install_requires=[
        'SquareMap3 >= 1.0.3',
        'wxPython',
    ],
    license="BSD",
    package_dir={
        'profile_viewer': 'profile_viewer',
    },
    packages=[
        'profile_viewer',
    ],
    options={
        'sdist': {
            'force_manifest': 1,
            'formats': ['gztar', 'zip'],
        },
    },
    zip_safe=False,
    entry_points={
        'gui_scripts': [
            'runsnake=profile_viewer.runsnake:main',
            'runsnakemem=profile_viewer.runsnake:meliaemain',
        ],
    },
    use_2to3=True,
    setup_requires=['setuptools_scm'],
    long_description_markdown_filename='README.md',
    classifiers=[
        """License :: OSI Approved :: BSD License""",
        """Programming Language :: Python""",
        """Topic :: Software Development :: Libraries :: Python Modules""",
        """Intended Audience :: Developers""",
    ],
    keywords='profile,gui,wxPython,squaremap',
    platforms=['Any'],
    install_package_data=True,
    use_scm_version=True)



"""Design-time __init__.py for resourcepackage

This is the scanning version of __init__.py for your
resource modules. You replace it with a blank or doc-only
init when ready to release.
"""
try:
	__file__
except NameError:
	pass
else:
	import os
	if os.path.splitext(os.path.basename( __file__ ))[0] == "__init__":
		try:
			from resourcepackage import package, defaultgenerators
			generators = defaultgenerators.generators.copy()
			
			### CUSTOMISATION POINT
			## import specialised generators here, such as for wxPython
			#from resourcepackage import wxgenerators
			#generators.update( wxgenerators.generators )
		except ImportError:
			pass
		else:
			package = package.Package(
				packageName = __name__,
				directory = os.path.dirname( os.path.abspath(__file__) ),
				generators = generators,
			)
			package.scan(
				### CUSTOMISATION POINT
				## force true -> always re-loads from external files, otherwise
				## only reloads if the file is newer than the generated .py file.
				# force = 1, 
			)
		



def macshim():
    """Shim to run 32-bit on 64-bit mac as a sub-process"""
    import subprocess, sys
    subprocess.call([
        sys.argv[0] + '32'
    ]+sys.argv[1:], 
        env={"VERSIONER_PYTHON_PREFER_32_BIT":"yes"}
    )
